#! /bin/sh
# demo.sh -- 1.48 -- 2020/09/17 11:42:28

TESTOR_DIR=${TESTOR_DIR:-..}

# cleanup generated files

if [ "$1" = -clean ] ; then
	# removal of files for jard_jeron_05
	rm -f jard_jeron_05_purpose.bcg
	rm -f jard_jeron_05*.ctg.bcg jard_jeron_05.tc.bcg
	rm -f jard_jeron_05_sut
	rm -f jard_jeron_05_sut?*.o jard_jeron_05_sut?.bcg
	# removal of files for the DES
	rm -f des_purpose_?*.bcg des_purpose_2_tgv.*.bcg
	rm -f des_product.*.bcg des_product_data.*.bcg
	rm -f des des.log
	# removal of common files
	rm -f sut.input sut.output
	exit
fi

# check the presence of CADP

if [ "$CADP" = "" ] ; then
	echo "$0: \$CADP should point to the installation directory of CADP"
	exit 1
elif [ ! -d $CADP ] ; then
	echo "$0: directory \$CADP does not exist"
	exit 1
fi

ARCH=`$CADP/com/arch`

# choice of the version of the DES example

if [ "$1" = -original-des ] ; then
	# original version with parallel cipher function
	# large state space, yielding prohibitively long execution times and
	# high memory requirements (except when using the LNT par operator to
	# indicate ACCEPT states)
	DES_VERSION=original
elif [ "$1" = -serialized-des ] ; then
	# simplified cipher function with serialized S-boxes
	# reasonably large state space, yielding long execution times for
	# testor (about 8 hours for the extraction of one test case) and
	# high memory requirements for TGV (more than 32 GB for purpose 1)
	# the table in [Marsso-Mateescu-Serwe-18] uses this version
	DES_VERSION=serialized
else
	# atomic cipher function, implemented by an LNT function
	# small state space, yielding short execution time (default option)
	DES_VERSION=atomic
fi

echo "-------------------------------------------------------------------------"
echo "example taken from [Jard-Jeron-05]"
echo "-------------------------------------------------------------------------"

echo "computing test purpose"
lnt.open -silent jard_jeron_05_purpose.lnt generator -rename tgv.rename jard_jeron_05_purpose.bcg

echo "extracting test case on the fly using testor"
lnt.open jard_jeron_05_spec.lnt $TESTOR_DIR/bin.$ARCH/testor.a -io jard_jeron_05.io jard_jeron_05_purpose.bcg jard_jeron_05.tc.bcg

echo
echo "compute complete test graph (synchronous product): jard_jeron_05.ctg.bcg"
lnt.open jard_jeron_05_spec.lnt $TESTOR_DIR/bin.$ARCH/testor.a -io jard_jeron_05.io -ctg jard_jeron_05_purpose.bcg jard_jeron_05.ctg.bcg

echo
echo "compute complete test graph (LNT parallel): jard_jeron_05_product.ctg.bcg"
lnt.open jard_jeron_05_product.lnt $TESTOR_DIR/bin.$ARCH/testor.a -io jard_jeron_05.io -product -accept $TESTOR_DIR/src/match_accept -refuse $TESTOR_DIR/src/match_refuse -ctg jard_jeron_05_product.ctg.bcg

echo
echo "checking equivalence of both complete test graphs"
bcg_cmp jard_jeron_05.ctg.bcg jard_jeron_05_product.ctg.bcg

echo
echo "checking inclusion of test case in the complete test graph"
bcg_open jard_jeron_05.ctg.bcg bisimulator -greater jard_jeron_05.tc.bcg |
grep -v '^bcg_open:'

echo
echo "generating executable implementation"

$CADP/src/com/cadp_cc jard_jeron_05_sut.c -o jard_jeron_05_sut

echo
echo "on-line testing of the SUT (derived from the model)"

mkfifo sut.input
mkfifo sut.output

# using the seed 42 leads to an Inconc verdict
./jard_jeron_05_sut -seed 0 > sut.output < sut.input &
# execution time measuring tools (e.g., memtime) write to the standard error;
# to avoid problems with the pipe connection, deactivate them during the test
(
	unset CADP_TIME
	lnt.open jard_jeron_05_spec.lnt $TESTOR_DIR/bin.$ARCH/testor.a -interactive -io jard_jeron_05.io jard_jeron_05_purpose.bcg < sut.output 2> sut.input
)

for SUT in sut1 sut2 sut3 ; do
	# note : testing jard_jeron_05_sut4 loops infinitely
	echo
	echo "on-line testing of jard_jeron_05_$SUT"

	bcg_io jard_jeron_05_$SUT.aut .bcg
	$TESTOR_DIR/bin.$ARCH/bcg_execute -io jard_jeron_05.io jard_jeron_05_$SUT > sut.output < sut.input &
	# reuse ./testor
	./testor -interactive -io jard_jeron_05.io jard_jeron_05_purpose.bcg < sut.output 2> sut.input
done

echo
echo "-------------------------------------------------------------------------"
echo "examples based on \$CADP/demos/demo_38 (DES)"
echo "-------------------------------------------------------------------------"

# copy the LNT specification from the CADP demo
cp $CADP/demos/demo_38/*.lnt .
chmod u+w DES.lnt CIPHER.lnt

# specify C names for 48-bit vectors
rm -f TYPES.lnt
sed -e 's/type BIT48 is/type BIT48 is !implementedby "ADT_BIT48" !printedby "ADT_PRINT_BIT48"/' $CADP/demos/demo_38/TYPES.lnt > TYPES.lnt

# change of the DES to make rendezvous on SUBKEY observable
rm -f DES.lnt
cat > DES.lnt << \EOF
module DES (CONTROLLER, DATA_PATH, KEY_PATH) is
process MAIN [CRYPT: CB, KEY, DATA, OUTPUT: C64, SUBKEY: C48] is
   hide CTRL_CL, CTRL_CR: CP, CTRL_SHIFT: CS, CTRL_DK, CTRL_CK: CP
   in
      par
         SUBKEY, CTRL_CL, CTRL_CR ->
            DATA_PATH [DATA, OUTPUT, SUBKEY, CTRL_CL, CTRL_CR]
      ||
         SUBKEY, CTRL_CK, CTRL_SHIFT, CTRL_DK ->
            KEY_PATH [KEY, SUBKEY, CTRL_SHIFT, CTRL_DK, CTRL_CK]
      ||
         CTRL_CL, CTRL_CR, CTRL_CK, CTRL_SHIFT, CTRL_DK ->
            CONTROLLER [CRYPT, CTRL_CL, CTRL_CR, CTRL_SHIFT, CTRL_DK, CTRL_CK]
      end par
   end hide
end process
end module
EOF

echo "configuring specification to use abstract bits"
sed -e 's/BIT_ABSTRACT/BIT/' BIT_ABSTRACT.lnt > BIT.lnt
# note that the use of abstract bits implies that all bits are 0, so that
# all constants of DES_CONSTANTS.lnt are the same

# simplification of the cipher function
if [ "$DES_VERSION" = atomic ] ; then
	# replacement of the parallel cipher function
	echo "replacing parallel cipher by an atomically executed function"
	rm -f CIPHER.lnt
	cp des_cipher.lnt CIPHER.lnt
	chmod u+w CIPHER.lnt
elif [ "$DES_VERSION" = serialized ] ; then
	# replacement of the parallel S-boxes by sequential ones
	echo "replacing parallel cipher by a sequential one"
	patch CIPHER.lnt < serialize_s_boxes
fi

# extraction of test cases with testor
# "-total" is necessary because the labels of the test purpose contain values
for PURPOSE in des_purpose_1 des_purpose_2 des_purpose_2_tgv
do
	echo
	echo "generating LTS for purpose \"$PURPOSE\""
	lnt.open -silent $PURPOSE.lnt generator -rename tgv.rename $PURPOSE.bcg
	echo
	echo "computing complete test graph using testor for purpose \"$PURPOSE\""
	lnt.open -silent DES.lnt $TESTOR_DIR/bin.$ARCH/testor.a -total -io des.io -ctg $PURPOSE.bcg $PURPOSE.ctg.bcg
	echo
	echo "computing test case on the fly using testor for purpose \"$PURPOSE\""
	lnt.open -silent DES.lnt $TESTOR_DIR/bin.$ARCH/testor.a -total -io des.io $PURPOSE.bcg $PURPOSE.tc.bcg
done

echo
echo "computing complete test graph using testor (LNT parallel)"
lnt.open -silent des_product.lnt $TESTOR_DIR/bin.$ARCH/testor.a -io des.io -product -accept $TESTOR_DIR/src/match_accept -refuse $TESTOR_DIR/src/match_refuse -ctg des_product.ctg.bcg

echo
echo "computing test case on the fly using testor (LNT parallel)"
lnt.open -silent des_product.lnt $TESTOR_DIR/bin.$ARCH/testor.a -io des.io -product -accept $TESTOR_DIR/src/match_accept -refuse $TESTOR_DIR/src/match_refuse des_product.tc.bcg

# extraction of test cases with TGV
for PURPOSE in des_purpose_1 des_purpose_2_tgv
do
	echo
	echo "computing complete test graph using tgv for purpose \"$PURPOSE\""
	lnt.open -silent DES.lnt tgv -self -io des.io -csg -output $PURPOSE.ctg.tgv.bcg $PURPOSE.bcg
	echo
	echo "computing test case on the fly using tgv for purpose \"$PURPOSE\""
	lnt.open -silent DES.lnt tgv -self -io des.io -output $PURPOSE.tc.tgv.bcg $PURPOSE.bcg
done
rm -f *.o

echo
echo "configuring specification to use concrete bits"
sed -e 's/BIT_CONCRETE/BIT/' BIT_CONCRETE.lnt > BIT.lnt

echo
echo "computing complete test graph using testor (LNT parallel)"
lnt.open -silent des_product_data.lnt $TESTOR_DIR/bin.$ARCH/testor.a -io des.io -product -accept $TESTOR_DIR/src/match_accept -refuse $TESTOR_DIR/src/match_refuse -ctg des_product_data.ctg.bcg
echo
echo "computing test case on the fly using testor (LNT parallel)"
lnt.open -silent des_product_data.lnt $TESTOR_DIR/bin.$ARCH/testor.a -io des.io -product -accept $TESTOR_DIR/src/match_accept -refuse $TESTOR_DIR/src/match_refuse des_product_data.tc.bcg

echo
echo "generating executable implementation"

lnt.open -silent DES.lnt -
caesar.adt -silent DES.lotos
caesar -silent -exec DES.lotos
$CADP/src/com/cadp_cc -I$CADP/incl -DCAESAR_KERNEL_DELAY=1 -DLNT DES.c gate_functions.c $CADP/src/exec_caesar/main.c -L$CADP/bin.`$CADP/com/arch` -lcaesar_base -lm -o des

echo
echo "on-line testing of the SUT for des_product_data"

if [ ! -p sut.input ] ; then
	mkfifo sut.input
fi
if [ ! -p sut.output ] ; then
	mkfifo sut.output
fi

./des des.log > sut.output < sut.input &
lnt.open -silent des_product_data.lnt $TESTOR_DIR/bin.$ARCH/testor.a -io des.io -product -accept $TESTOR_DIR/src/match_accept -refuse $TESTOR_DIR/src/match_refuse -interactive < sut.output 2> sut.input

# removal of the LNT specification of the DES
rm -f BIT.lnt BIT_ABSTRACT.lnt BIT_CONCRETE.lnt CHANNELS.lnt CIPHER.lnt
rm -f CONTROLLER.lnt DATA_PATH.lnt DES.lnt DES_SAMPLE.lnt KEY_PATH.lnt
rm -f PERMUTATION_FUNCTIONS.lnt S_BOX_FUNCTIONS.lnt TYPES.lnt property_4.lnt

rm -f *.t *.f *.lotos *.lib *.h *.o
rm -f `ls -1 *.c | grep -v gate_functions.c | grep -v jard_jeron_05_sut.c`
rm -f bisimulator generator reductor testor tgv
