/******************************************************************************
 *             D A T A    E N C R Y P T I O N    S T A N D A R D
 *-----------------------------------------------------------------------------
 *   INRIA - Unite de Recherche Rhone-Alpes
 *   655, avenue de l'Europe
 *   38330 Montbonnot Saint Martin
 *   FRANCE
 *-----------------------------------------------------------------------------
 *   Module		:	gate_functions.c
 *   Auteur		:	Wendelin SERWE
 *   Version		:	1.20
 *   Date		: 	2015/09/11 12:26:43
 *****************************************************************************
 * patched by Wendelin SERWE -- 2020/09/04 17:10:37
 * - to handle also the SUBKEY gate
 * - to parse inputs in the same way they are printed by default
 * - to avoid blocking on inputs
 *****************************************************************************/

/*
 * Gate functions for the DES prototype generated using EXEC/CAESAR
 * 
 * The DES prototype reads its inputs from stdin and prints its results to
 * stdout. Each rendezvous corresponds to one line of input (respectively,
 * output), following the syntax of LOTOS, i.e., "<GATE> !<OFFER>". For
 * convenience, 64-bit vectors are represented by a sequence of sixteen
 * hexadecimal digits. The prototype can also write its execution trace in
 * the SEQ format to a log file.
 */

#include <fcntl.h>

/* ------------------------------------------------------------------------- */

#define CAESAR_ADT_INTERFACE

#if defined (LOTOS)
#include "DES_CONCRETE.h"
#elif defined (LNT)
#include "DES.h"
#else
#error "compile with either -DLOTOS or -DLNT"
#endif

/* ------------------------------------------------------------------------- */

#include "caesar_kernel.h"

extern void CAESAR_KERNEL_EXIT ();

#define streq(S1,S2) (strcmp ((S1), (S2)) == 0)

/* ------------------------------------------------------------------------- */

#define CHAR_TO_BIT(C) (((C) == '0') ? BIT_ZERO () : BIT_ONE ())

/* ------------------------------------------------------------------------- */

/* buffer for the gate of the rendezvous in the current line of stdin */

static char CURRENT_GATE[6];

/* buffer for the offer of the rendezvous in the current line of stdin */

static char CURRENT_OFFER[65];

/*
 * if equal to zero, CURRENT_GATE and CURRENT_OFFER contain the gate (resp.
 * OFFER) of the rendezvous in the current line of stdin
 */

static char CURRENT_LINE_NOT_READ = 1;

/* ------------------------------------------------------------------------- */

/*
 * auxiliary function to check whether the current line of stdin corresponds
 * to a rendezvous on gate GATE
 */

int  PARSE_INPUT (GATE, OFFER)
char *GATE;
void *OFFER;			/* a caster */
{
     int  N;

     if (feof (stdin))
	  return (0);

     if (CURRENT_LINE_NOT_READ) {
	  /* check whether there is something to read on standard input */
	  fcntl (STDIN_FILENO, F_SETFL, fcntl (STDIN_FILENO, F_GETFL, 0) | O_NONBLOCK);

	  /* read a line from stdin */
	  N = scanf ("CRYPT !%5[AEFLRSTU]\n", CURRENT_OFFER);
	  if (N == EOF)
	       return (0);
	  if (N == 1) {
	       strcpy (CURRENT_GATE, "CRYPT");
	  } else {
	       N = scanf ("%5s !MK_64 (%1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01], %1[01])\n",
			  CURRENT_GATE,
			  CURRENT_OFFER,
			  CURRENT_OFFER + 1,
			  CURRENT_OFFER + 2,
			  CURRENT_OFFER + 3,
			  CURRENT_OFFER + 4,
			  CURRENT_OFFER + 5,
			  CURRENT_OFFER + 6,
			  CURRENT_OFFER + 7,
			  CURRENT_OFFER + 8,
			  CURRENT_OFFER + 9,
			  CURRENT_OFFER + 10,
			  CURRENT_OFFER + 11,
			  CURRENT_OFFER + 12,
			  CURRENT_OFFER + 13,
			  CURRENT_OFFER + 14,
			  CURRENT_OFFER + 15,
			  CURRENT_OFFER + 16,
			  CURRENT_OFFER + 17,
			  CURRENT_OFFER + 18,
			  CURRENT_OFFER + 19,
			  CURRENT_OFFER + 20,
			  CURRENT_OFFER + 21,
			  CURRENT_OFFER + 22,
			  CURRENT_OFFER + 23,
			  CURRENT_OFFER + 24,
			  CURRENT_OFFER + 25,
			  CURRENT_OFFER + 26,
			  CURRENT_OFFER + 27,
			  CURRENT_OFFER + 28,
			  CURRENT_OFFER + 29,
			  CURRENT_OFFER + 30,
			  CURRENT_OFFER + 31,
			  CURRENT_OFFER + 32,
			  CURRENT_OFFER + 33,
			  CURRENT_OFFER + 34,
			  CURRENT_OFFER + 35,
			  CURRENT_OFFER + 36,
			  CURRENT_OFFER + 37,
			  CURRENT_OFFER + 38,
			  CURRENT_OFFER + 39,
			  CURRENT_OFFER + 40,
			  CURRENT_OFFER + 41,
			  CURRENT_OFFER + 42,
			  CURRENT_OFFER + 43,
			  CURRENT_OFFER + 44,
			  CURRENT_OFFER + 45,
			  CURRENT_OFFER + 46,
			  CURRENT_OFFER + 47,
			  CURRENT_OFFER + 48,
			  CURRENT_OFFER + 49,
			  CURRENT_OFFER + 50,
			  CURRENT_OFFER + 51,
			  CURRENT_OFFER + 52,
			  CURRENT_OFFER + 53,
			  CURRENT_OFFER + 54,
			  CURRENT_OFFER + 55,
			  CURRENT_OFFER + 56,
			  CURRENT_OFFER + 57,
			  CURRENT_OFFER + 58,
			  CURRENT_OFFER + 59,
			  CURRENT_OFFER + 60,
			  CURRENT_OFFER + 61,
			  CURRENT_OFFER + 62,
			  CURRENT_OFFER + 63);
	       if (N == EOF)
		    return (0);
	       if (N != 65) {
		    if (streq (CURRENT_GATE, "Exit")) {
			 /*
			  * the interactive tester runs under memtime and has
			  * terminated its execution
			  */
			 CAESAR_KERNEL_TERM ();
			 exit (0);
			 /* NOTREACHED */
		    }
		    fprintf (stderr, "\ninput error\n");
		    fprintf (stderr, "CURRENT_GATE == \"%s\"\n", CURRENT_GATE);
		    if (streq (CURRENT_GATE, "CRYPT")) {
			 fprintf (stderr, "CURRENT_OFFER == \"%s\"\n", CURRENT_OFFER);
		    }
		    CAESAR_KERNEL_EXIT ("incorrect input syntax\n");
		    /* NOTREACHED */
	       }
	  }
	  CURRENT_LINE_NOT_READ = 0;
     }
     if (!streq (CURRENT_GATE, GATE))
	  return (0);
     CURRENT_LINE_NOT_READ = 1;

     if (streq (GATE, "CRYPT")) {
	  if (streq (CURRENT_OFFER, "FALSE")) {
	       *((ADT_BOOL *) OFFER) = ADT_FALSE ();
	  } else if (streq (CURRENT_OFFER, "TRUE")) {
	       *((ADT_BOOL *) OFFER) = ADT_TRUE ();
	  } else {
	       CAESAR_KERNEL_EXIT ("incorrect offer \"%s\" for gate %s\n",
				   CURRENT_OFFER, GATE);
	       /* NOTREACHED */
	  }
     } else if (streq (GATE, "DATA") || streq (GATE, "KEY")) {
	  *((ADT_BIT64 *) OFFER) =
	       MK_64 (CHAR_TO_BIT (CURRENT_OFFER[0]),
		      CHAR_TO_BIT (CURRENT_OFFER[1]),
		      CHAR_TO_BIT (CURRENT_OFFER[2]),
		      CHAR_TO_BIT (CURRENT_OFFER[3]),
		      CHAR_TO_BIT (CURRENT_OFFER[4]),
		      CHAR_TO_BIT (CURRENT_OFFER[5]),
		      CHAR_TO_BIT (CURRENT_OFFER[6]),
		      CHAR_TO_BIT (CURRENT_OFFER[7]),
		      CHAR_TO_BIT (CURRENT_OFFER[8]),
		      CHAR_TO_BIT (CURRENT_OFFER[9]),
		      CHAR_TO_BIT (CURRENT_OFFER[10]),
		      CHAR_TO_BIT (CURRENT_OFFER[11]),
		      CHAR_TO_BIT (CURRENT_OFFER[12]),
		      CHAR_TO_BIT (CURRENT_OFFER[13]),
		      CHAR_TO_BIT (CURRENT_OFFER[14]),
		      CHAR_TO_BIT (CURRENT_OFFER[15]),
		      CHAR_TO_BIT (CURRENT_OFFER[16]),
		      CHAR_TO_BIT (CURRENT_OFFER[17]),
		      CHAR_TO_BIT (CURRENT_OFFER[18]),
		      CHAR_TO_BIT (CURRENT_OFFER[19]),
		      CHAR_TO_BIT (CURRENT_OFFER[20]),
		      CHAR_TO_BIT (CURRENT_OFFER[21]),
		      CHAR_TO_BIT (CURRENT_OFFER[22]),
		      CHAR_TO_BIT (CURRENT_OFFER[23]),
		      CHAR_TO_BIT (CURRENT_OFFER[24]),
		      CHAR_TO_BIT (CURRENT_OFFER[25]),
		      CHAR_TO_BIT (CURRENT_OFFER[26]),
		      CHAR_TO_BIT (CURRENT_OFFER[27]),
		      CHAR_TO_BIT (CURRENT_OFFER[28]),
		      CHAR_TO_BIT (CURRENT_OFFER[29]),
		      CHAR_TO_BIT (CURRENT_OFFER[30]),
		      CHAR_TO_BIT (CURRENT_OFFER[31]),
		      CHAR_TO_BIT (CURRENT_OFFER[32]),
		      CHAR_TO_BIT (CURRENT_OFFER[33]),
		      CHAR_TO_BIT (CURRENT_OFFER[34]),
		      CHAR_TO_BIT (CURRENT_OFFER[35]),
		      CHAR_TO_BIT (CURRENT_OFFER[36]),
		      CHAR_TO_BIT (CURRENT_OFFER[37]),
		      CHAR_TO_BIT (CURRENT_OFFER[38]),
		      CHAR_TO_BIT (CURRENT_OFFER[39]),
		      CHAR_TO_BIT (CURRENT_OFFER[40]),
		      CHAR_TO_BIT (CURRENT_OFFER[41]),
		      CHAR_TO_BIT (CURRENT_OFFER[42]),
		      CHAR_TO_BIT (CURRENT_OFFER[43]),
		      CHAR_TO_BIT (CURRENT_OFFER[44]),
		      CHAR_TO_BIT (CURRENT_OFFER[45]),
		      CHAR_TO_BIT (CURRENT_OFFER[46]),
		      CHAR_TO_BIT (CURRENT_OFFER[47]),
		      CHAR_TO_BIT (CURRENT_OFFER[48]),
		      CHAR_TO_BIT (CURRENT_OFFER[49]),
		      CHAR_TO_BIT (CURRENT_OFFER[50]),
		      CHAR_TO_BIT (CURRENT_OFFER[51]),
		      CHAR_TO_BIT (CURRENT_OFFER[52]),
		      CHAR_TO_BIT (CURRENT_OFFER[53]),
		      CHAR_TO_BIT (CURRENT_OFFER[54]),
		      CHAR_TO_BIT (CURRENT_OFFER[55]),
		      CHAR_TO_BIT (CURRENT_OFFER[56]),
		      CHAR_TO_BIT (CURRENT_OFFER[57]),
		      CHAR_TO_BIT (CURRENT_OFFER[58]),
		      CHAR_TO_BIT (CURRENT_OFFER[59]),
		      CHAR_TO_BIT (CURRENT_OFFER[60]),
		      CHAR_TO_BIT (CURRENT_OFFER[61]),
		      CHAR_TO_BIT (CURRENT_OFFER[62]),
		      CHAR_TO_BIT (CURRENT_OFFER[63]));
     } else {
	  CAESAR_KERNEL_EXIT ("unknown gate \"%s\"\n", CURRENT_GATE);
	  /* NOTREACHED */
     }
     return (1);
}

/*---------------------------------------------------------------------------*/

int  CRYPT (INPUT_MODE, BOOL, OFFER, EOL)
CAESAR_KERNEL_OFFER INPUT_MODE;
char *BOOL;
ADT_BOOL *OFFER;
CAESAR_KERNEL_OFFER EOL;
{
     int  CAESAR_STATUS;

     CAESAR_KERNEL_ASSERT_INPUT (INPUT_MODE);
     CAESAR_KERNEL_ASSERT_TYPE (BOOL, "ADT_BOOL");
     CAESAR_KERNEL_ASSERT_EOL (EOL);

     CAESAR_STATUS = PARSE_INPUT ("CRYPT", OFFER);

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (INPUT_MODE, BOOL, *OFFER, ADT_PRINT_BOOL, CAESAR_STATUS);
     CAESAR_KERNEL_LOG_RESULT (CAESAR_STATUS);

     return (CAESAR_STATUS);
}

/* ------------------------------------------------------------------------- */

int  DATA (INPUT_MODE, BIT64, OFFER, EOL)
CAESAR_KERNEL_OFFER INPUT_MODE;
char *BIT64;
ADT_BIT64 *OFFER;
CAESAR_KERNEL_OFFER EOL;
{
     int  CAESAR_STATUS;

     CAESAR_KERNEL_ASSERT_INPUT (INPUT_MODE);
     CAESAR_KERNEL_ASSERT_TYPE (BIT64, "ADT_BIT64");
     CAESAR_KERNEL_ASSERT_EOL (EOL);

     CAESAR_STATUS = PARSE_INPUT ("DATA", OFFER);

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (INPUT_MODE, BIT64, *OFFER, ADT_PRINT_BIT64, CAESAR_STATUS);
     CAESAR_KERNEL_LOG_RESULT (CAESAR_STATUS);

     return (CAESAR_STATUS);
}

/* ------------------------------------------------------------------------- */

int  KEY (INPUT_MODE, BIT64, OFFER, EOL)
CAESAR_KERNEL_OFFER INPUT_MODE;
char *BIT64;
ADT_BIT64 *OFFER;
CAESAR_KERNEL_OFFER EOL;
{
     int  CAESAR_STATUS;

     CAESAR_KERNEL_ASSERT_INPUT (INPUT_MODE);
     CAESAR_KERNEL_ASSERT_TYPE (BIT64, "ADT_BIT64");
     CAESAR_KERNEL_ASSERT_EOL (EOL);

     CAESAR_STATUS = PARSE_INPUT ("KEY", OFFER);

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (INPUT_MODE, BIT64, *OFFER, ADT_PRINT_BIT64, CAESAR_STATUS);
     CAESAR_KERNEL_LOG_RESULT (CAESAR_STATUS);

     return (CAESAR_STATUS);
}

/* ------------------------------------------------------------------------- */

int  OUTPUT (OUTPUT_MODE, BIT64, OFFER, EOL)
CAESAR_KERNEL_OFFER OUTPUT_MODE;
char *BIT64;
ADT_BIT64 OFFER;
CAESAR_KERNEL_OFFER EOL;
{
     CAESAR_KERNEL_ASSERT_OUTPUT (OUTPUT_MODE);
     CAESAR_KERNEL_ASSERT_TYPE (BIT64, "ADT_BIT64");
     CAESAR_KERNEL_ASSERT_EOL (EOL);

     fprintf (stdout, "OUTPUT !");
     ADT_PRINT_BIT64 (stdout, OFFER);
     fprintf (stdout, "\n");
     fflush (stdout);

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (OUTPUT_MODE, BIT64, OFFER, ADT_PRINT_BIT64, 1);
     CAESAR_KERNEL_LOG_RESULT (1);

     return (1);
}

/* ------------------------------------------------------------------------- */

int  SUBKEY (OUTPUT_MODE, BIT48, OFFER, EOL)
CAESAR_KERNEL_OFFER OUTPUT_MODE;
char *BIT48;
ADT_BIT48 OFFER;
CAESAR_KERNEL_OFFER EOL;
{
     CAESAR_KERNEL_ASSERT_OUTPUT (OUTPUT_MODE);
     CAESAR_KERNEL_ASSERT_TYPE (BIT48, "ADT_BIT48");
     CAESAR_KERNEL_ASSERT_EOL (EOL);

     fprintf (stdout, "SUBKEY !");
     ADT_PRINT_BIT48 (stdout, OFFER);
     fprintf (stdout, "\n");
     fflush (stdout);

     CAESAR_KERNEL_LOG_GATE (__func__);
     /* LINTED */
     CAESAR_KERNEL_LOG_OFFER (OUTPUT_MODE, BIT48, OFFER, ADT_PRINT_BIT48, 1);
     CAESAR_KERNEL_LOG_RESULT (1);

     return (1);
}

/* ------------------------------------------------------------------------- */
